#coding: utf-8
#
#Copyright (c) 2010 Ukyo http://hujimi.seesaa.net/
#Licensed under the MIT license: http://www.opensource.org/licenses/mit-license.php

from google.appengine.ext import webapp
from google.appengine.ext.webapp.util import run_wsgi_app
from google.appengine.api import memcache
from models import BBSThread, BBSComment, BBSThreadsLength
from django.utils import feedgenerator

import threading
import time
import uuid
import math
import cgi
import re

TEMPLATE_HTML = u'''
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN">

<html>
    <head>
        <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
        <title>%s%s</title>
        <link href="/css/main.css" rel="stylesheet" type="text/css" />
        <link href="/css/prettify.css" rel="stylesheet" type="text/css" />
        <link rel="alternate" type="application/rss+xml" title="RSS" href="%s" />
    </head>
    <body>
        <div id="container">
        <div id="screen">
            <h1><a href="%s">%s</a></h1>
            %s
            <div id="footer">
                <div id="copyright">Copyright (c) 2010. <a href="http://hujimi.seesaa.net">Ukyo</a></div>
            </div>
        </div>
        </div>
        <script type="text/javascript" src="/js/prettify.js"></script>
        <script type="text/javascript">try{
            window.addEventListener("load",prettyPrint,false);
        }catch(e){
            window.attachEvent("onload",prettyPrint);
        }
        </script>
    </body>
</html>
'''


'''設定'''
PAGE_LENGTH = 20
THREADS_LENGTH_MAX = 100
COMMENTS_LENGTH_MAX = 1000
RSS_LENGTH = 100

APP_URL = 'http://web-board.appspot.com'
#APP_URL = 'http://localhost:8080'
APP_TITLE = 'BBS for Google App Engine'
PATH_MAIN = '/'
PATH_THREAD = '/thread'
PATH_COMMENT = '/comment'
PATH_THREAD_CREATE = '/createthread'
PATH_RSS = '/rss'


'''memcacheからデータを読み込む際に使うキー'''
MEMCACHE_KEY_THREADS_LENGTH = 'bbs_thread_length'
MEMCACHE_KEY_THREADS = 'bbs_threads'
MEMCACHE_KEY_COMMENTS = 'bbs_comments'
MEMCACHE_KEY_FEED = 'feed'


'''decorate_textで使用するパターン'''

'''2chのアンカーリンクのようなものを作る'''
pattern_2ch_ancher = re.compile('&gt;&gt;([0-9]+)')
'''http://で始まるテキストをクリック可能にする'''
pattern_link = re.compile('(http://[A-Za-z0-9\'~+\-=_.,/%\?!;:@#\*&\(\)]+)')
'''ソースコードを見栄え良く表示'''
pattern_source = re.compile('&gt;\|\|(.*?)\|\|&lt;')

def decorate_text(text):
    '''コメント本文を装飾して返します。
    
    Args:
        text: コメント本文。
    Returns:
        装飾されたコメント本文。
    '''
    text = cgi.escape(text)
    
    text = pattern_2ch_ancher.sub(r'<a href="#\1">&gt;&gt;\1</a>', text)
    text = pattern_link.sub(r'<a href="\1" target="_blank">\1</a>', text)
    text = text.replace('\r\n', '<br />')
    text = pattern_source.sub(r'<pre class="prettyprint">\1</pre>', text)
    
    return text


def build_html(content, thread_title='', thread_key=''):
    '''htmlを生成して返します。
    
    Args:
        content: コンテンツ。掲示板スレッド一覧やコメント一覧など。
        thread_title: 掲示板スレッドのタイトル。省略可。
        thread_key: 掲示板スレッドのキー。省略可。
    Returns:
        生成されたhtmlを返す。
    '''
    if thread_title:
        thread_title += ' - '
    if thread_key:
        thread_key = '?q=' + thread_key
        
    return TEMPLATE_HTML % (thread_title,
                            APP_TITLE,
                            PATH_RSS + thread_key,
                            PATH_MAIN,
                            APP_TITLE,
                            content)
    
    
def get_bbs_threads_length():
    '''データストアに保存されている掲示板スレッド数を返します。
    
    return:
        bbs_threadMax: 掲示板スレッド数(integer)。
    '''
    bbs_threads_length = memcache.get(MEMCACHE_KEY_THREADS_LENGTH)
    if bbs_threads_length is None:
        bbs_threads_length = BBSThreadsLength.all().get()
        if bbs_threads_length is None:
            bbs_threads_length = BBSThreadsLength(n=0)
            bbs_threads_length.put()
        memcache.add(MEMCACHE_KEY_THREADS_LENGTH, bbs_threads_length)
    return bbs_threads_length


def get_bbs_threads(offset=0):
    '''データストアに保存されている掲示板スレッドのうち、
    更新時間の降順で指定されたページ分だけ読み込んで返します。
    Args:
        offset: オフセット。PAGE_LENGTH分だけずらします。
    Returns:
        掲示板スレッドの一覧(html)。
    '''
    if offset:
        bbs_threads_ = BBSThread.all().order("-updated_at").fetch(PAGE_LENGTH, PAGE_LENGTH*offset)
        bbs_threads = []
        for thread in bbs_threads_:
            bbs_threads.append(build_thread(thread))
        bbs_threads = ''.join(bbs_threads)
    else:
        bbs_threads = memcache.get(MEMCACHE_KEY_THREADS)
        if bbs_threads is None:
            bbs_threads_ = BBSThread.all().order("-updated_at").fetch(PAGE_LENGTH)
            bbs_threads = []
            for thread in bbs_threads_:
                bbs_threads.append(build_thread(thread))
            bbs_threads = ''.join(bbs_threads)
            memcache.add(MEMCACHE_KEY_THREADS, bbs_threads)
    return bbs_threads


def get_bbs_thread(thread_key):
    '''データストアからキーに該当する掲示板スレッドを検索して返します。
    
    Args:
        thread_key: 掲示板スレッドのユニークキー。
    Returns:
        掲示板スレッド(BBSThread)。
    '''
    bbs_thread = memcache.get(thread_key)
    if bbs_thread is None:
        bbs_thread = BBSThread.get_by_key_name(thread_key)
        memcache.add(thread_key, bbs_thread)
    return bbs_thread


def get_bbs_comments(thread_key):
    '''データストアからキーに該当するコメント一覧を検索して返します。
    
    Args:
        thread_key: 掲示板スレッドのユニークキー。
    Returns:
        コメント一覧(html)。
    '''
    bbs_comments = memcache.get(MEMCACHE_KEY_COMMENTS + thread_key)

    if bbs_comments is None:
        bbs_comments_ = BBSComment.all().filter("thread_key", thread_key)
        bbs_comments = []
        count = 1
        for comment in bbs_comments_:
            bbs_comments.append(build_comment(count, comment))
            count += 1
        bbs_comments = ''.join(bbs_comments)
        memcache.add(MEMCACHE_KEY_COMMENTS + thread_key, bbs_comments)
    return bbs_comments


def build_error_page(messages):
    '''エラーメッセージのリストからエラーページ(html)を生成して返します
    
    Args:
        messages: エラーメッセージ(string)のリスト。
    Returns:
        生成されたエラーページ(html)。
    '''
    content = []
    content.append('<ul id="error-messages">')
    for message in messages: content.append('<li>%s is required</li>' % message)
    content.append('</ul>')
    return build_html(''.join(content), 'Error')


def build_comment(count, comment):
    '''コメント(html)を生成して返します。
    
    Args:
        count: コメントが何番目かを表す数字。
        comment: BBSComment。
    Returns:
        生成されたコメント(html)。
    '''
    return '''
            <a name="%d"></a>
            <div class="comment">
                <dl class="comment-header">
                    <dt>No.</dt><dd>%d</dd>
                    <dt>auther</dt><dd>%s</dd>
                    <dt>created at</dt><dd>%s</dd>
                </dl>
                <div class="comment-body">%s</div>
            </div>''' % (count,
                         count,
                         comment.auther,
                         comment.created_at.strftime("%Y-%m-%d %H:%M:%S"),
                         decorate_text(comment.text))


def build_thread(thread):
    '''掲示板スレッド(html)を生成して返します。
    
    Args:
        thread: BBSThread。
    Returns:
        生成された掲示板スレッド(html)。
    '''
    return '''
            <tr>
                <td class="thread-title"><a href="%s?q=%s">%s</td>
                <td class="thread-auther">%s</td>
                <td class="thread-comments">%d</td>
                <td class="thread-updated">%s</td>
                <td class="thread-created">%s</td>
            </tr>''' % (PATH_THREAD,
                        thread.key().name(),
                        thread.title,
                        thread.auther,
                        thread.length,
                        thread.updated_at.strftime("%Y-%m-%d %H:%M:%S"),
                        thread.created_at.strftime("%Y-%m-%d %H:%M:%S"))


def get_feed(thread_key=''):
    '''RSSのフィードを返します。
    
    Args:
        thread_key: 
    Returns:
        
    '''
    feed = memcache.get(MEMCACHE_KEY_FEED + thread_key)
    if feed is None:
        feed = feedgenerator.Rss201rev2Feed(
            title=APP_TITLE,
            link=APP_URL + PATH_RSS,
            description=APP_TITLE,
            language=u'ja')
        
        if thread_key:
            bbs_comments = BBSComment.all().order("-created_at").filter("thread_key", thread_key).fetch(RSS_LENGTH)
        else:
            bbs_comments = BBSComment.all().order("-created_at").fetch(RSS_LENGTH)
            
        for comment in bbs_comments:
            feed.add_item(title=get_bbs_thread(comment.thread_key).title,
                          link=APP_URL + PATH_THREAD + '?q=' + comment.thread_key,
                          description=decorate_text(comment.text))
            
        memcache.add(MEMCACHE_KEY_FEED + thread_key, feed)
    return feed


class MainPage(webapp.RequestHandler):
    '''メイン画面'''
    
    def get(self):
        '''主にスレッド一覧とスレッド作成フォームを表示します'''
        
        page = self.request.get("page")
        try: page = int(page)
        except: page = 0
        if page < 0: page = 0
        
        bbs_threads = get_bbs_threads(page)
        
        bbs_threads_length = get_bbs_threads_length()
        
        content = []
        
        content.append('''
        <div id="threads"><table>
        <tr>
            <th>Title</th>
            <th>Auther</th>
            <th>Comments</th>
            <th>Updated at</th>
            <th>Created at</th>
        </tr>''')
        content.append(bbs_threads)
        content.append('</table>')
        
        #navi
        content.append('<ul id="nav">')
        if page != 0:
            content.append('<li><a href="/?page=%d">Prev</a></li>' % (page - 1))
        else:
            content.append('<li>Prev</li>')
        
        content.append('<li>Page %d of %d</li>' % (page + 1,
                                                   int(math.floor((bbs_threads_length.n - 1) / PAGE_LENGTH) + 1)))
        
        if bbs_threads_length.n > PAGE_LENGTH * (page + 1):
            content.append('<li><a href="%s?page=%d">Next</a></li>' % (PATH_MAIN, page + 1))
        else:
            content.append('<li>Next</li>')
        content.append('</ul>')
        
        content.append('</div>')
        
        #form
        content.append('''
        <div id="new-thread">
            <form method="post" action="%s">
                <h2>Create a new thread</h2>
                <label>Title</label>
                <input type="text" name="title" maxlength="25">
                <label>Auther</label>
                <input type="text" name="auther" maxlength="10">
                <label>Text</label>
                <textarea name="text"></textarea>
                <button type="submit">Create</button>
            </form>
        </div>''' % PATH_THREAD_CREATE)
        
        self.response.out.write(build_html(''.join(content)))

            
class BBSThreadPage(webapp.RequestHandler):
    '''スレッド画面'''
    
    def get(self):
        '''スレッドに対応したコメントとコメント投稿フォームを表示します'''
        
        thread_key = self.request.get("q")
        
        bbs_comments = get_bbs_comments(thread_key)
        bbs_thread = get_bbs_thread(thread_key)
        
        content = []

        content.append('<div id="comments">')
        content.append(bbs_comments)
        content.append('</div>')
        if bbs_thread.length < COMMENTS_LENGTH_MAX:
            content.append('''
            <a name="new-comment"></a>
            <div id="new-comment">
                <form method="post" action="%s">
                    <h2>Submit a new comment</h2>
                    <input type="hidden" name="key" value="%s">
                    <label>Auther</label>
                    <input type="text" name="auther" maxlength="10">
                    <label>Text</label>
                    <textarea name="text"></textarea>
                    <button type="submit">Submit</button>
                </form>
            </div>''' % (PATH_COMMENT, thread_key))
        else:
            content.append('<p>This thread has 1000 comments. Create a new thread.</p>')
    
        content.append('<div id="back-to-mainpage"><a href="%s">Back to Main Page</a></div>' % PATH_MAIN)

        self.response.out.write(build_html(''.join(content), bbs_thread.title, thread_key))

    
class SafeThreadRequestHandler(webapp.RequestHandler, threading.Thread):
    '''スレッドセーフなリクエストハンドラ'''
    
    mutex = threading.Semaphore(1)
    
    def __init__(self):
        webapp.RequestHandler.__init__(self)
        threading.Thread.__init__(self)


class BBSThreadCreater(SafeThreadRequestHandler):
    '''掲示板スレッド作成クラス'''
    
    def post(self):
        '''Args(POSTで送られてきたデータ)をもとに、掲示板スレッドを作成します
        
        Args:
            title: 掲示板スレッドのタイトル(必須項目)
            auther: 作成者名(必須項目)
            text: 掲示板スレッドの最初のコメント本文(必須項目)
        '''
        
        title = cgi.escape(self.request.get("title"))
        auther = cgi.escape(self.request.get("auther"))
        text = self.request.get("text")
        
        empty = False
        messages = []
        if title == '':
            messages.append('Title')
            empty = True
        if auther == '':
            messages.append('Auther')
            empty = True
        if text == '':
            messages.append('Text')
            empty = True
        if empty:
            self.response.out.write(build_error_page(messages))
            return
        
        self.mutex.acquire()
        bbs_thread = BBSThread(key_name=str(time.time()) + str(uuid.uuid4()))
        bbs_thread.title = title
        bbs_thread.auther = auther
        bbs_thread.length = 1
        bbs_thread.put()
        memcache.delete(MEMCACHE_KEY_THREADS)
        
        bbs_comment = BBSComment()
        bbs_comment.auther = auther
        bbs_comment.text = text
        bbs_comment.thread_key = bbs_thread.key().name()
        bbs_comment.put()
        
        bbs_threads_length = get_bbs_threads_length()
        if bbs_threads_length.n > THREADS_LENGTH_MAX:
            BBSThread.all().order("updated_at").get().delete()
        else:
            bbs_threads_length.n += 1
            bbs_threads_length.put()
            memcache.set(MEMCACHE_KEY_THREADS_LENGTH, bbs_threads_length)
        
        memcache.delete(MEMCACHE_KEY_FEED)
        
        self.redirect(PATH_MAIN)
        self.mutex.release()
    

class BBSCommentCreater(SafeThreadRequestHandler):
    '''コメント作成クラス'''
    
    def post(self):
        '''Args(POSTで送られてきたデータ)をもとに、コメントを作成します
        
        Args:
            key: 対応する掲示板スレッドのキー(hidden属性で送られる)
            auther: 作成者名(必須項目)
            text: コメント本文(必須項目)
        '''
        
        thread_key = self.request.get("key")
        auther = cgi.escape(self.request.get("auther"))
        text = self.request.get("text")
        
        empty = False
        messages = []
        if auther == '':
            messages.append('Auther')
            empty = True
        if text == '':
            messages.append('Text')
            empty = True
        if empty:
            self.response.out.write(build_error_page(messages))
            return
        
        self.mutex.acquire()
        bbs_thread = get_bbs_thread(thread_key)
        bbs_thread.length += 1
        bbs_thread.put()
        
        bbs_comment = BBSComment()
        bbs_comment.thread_key = thread_key
        bbs_comment.auther = auther
        bbs_comment.text = text
        bbs_comments = get_bbs_comments(thread_key)
        bbs_comments += build_comment(bbs_thread.length, bbs_comment)
        bbs_comment.put()
        
        memcache.set(MEMCACHE_KEY_COMMENTS + thread_key, bbs_comments)
        memcache.delete(thread_key)
        memcache.delete(MEMCACHE_KEY_THREADS)
        memcache.delete(MEMCACHE_KEY_FEED)
        memcache.delete(MEMCACHE_KEY_FEED + '?q=' + thread_key)
        
        self.redirect(PATH_THREAD + '?q=' + thread_key + '#new-comment')
        self.mutex.release()


class RSSPage(webapp.RequestHandler):
    
    def get(self):
        thread_key = self.request.get("q")
        self.response.out.write(get_feed(thread_key).writeString('utf-8'))

application = webapp.WSGIApplication([(PATH_MAIN, MainPage),
                                      (PATH_THREAD, BBSThreadPage),
                                      (PATH_COMMENT, BBSCommentCreater),
                                      (PATH_THREAD_CREATE, BBSThreadCreater),
                                      (PATH_RSS, RSSPage)
                                      ], debug=True)


def main():
    run_wsgi_app(application)

if __name__ == "__main__":
    main()

